import datetime
import os
from ..tools import *  # Module under test


def test_sys_context_env_path():
    import os
    assert splice("[[SYS.ENV.PATH]]") == os.getenv("PATH")


def test_sys_uuid_length():
    assert len(splice("[[SYS.UUID]]")) == 32

# Test setting and reading environment variables
def test_env():
    import os
    os.environ["FOO"] = "BAR"
    assert splice("[[SYS.ENV.FOO]]") == "BAR"


# Check that SYS.ENV is a dictionary with PATH
def test_get_env():
    assert splice("[[SYS.ENV.PATH]]") == os.environ["PATH"]


# Test SYS.NOW produces correct year
def test_get_now():
    assert splice("[[SYS.NOW.YEAR]]") == str(datetime.datetime.now().year)


# Test SYS.PID produces correct pid
def test_get_pid():
    assert splice("[[SYS.PID]]") == str(os.getpid())


# Text getting the correct python executable
def test_get_python():
    import sys
    assert splice("[[SYS.PYTHON]]") == sys.executable



from ..tools import *  # Module under test
from ..contexts import context


def test_join_list():
    with context(LIST=["Hello", "World"]):
        assert splice("[[LIST.JOIN=', ']]") == "Hello, World"

# Test that a string has no sigils with spool
def test_no_sigils_in_big_string():
    with context(HELLO="Hello", WORLD="World"):
        assert not any(spool("Hello World"))

# Test adding a custom function to context
def test_custom_function():
    def custom_func():
        return "Hello World"

    with context(CUSTOM=custom_func):
        assert splice("[[CUSTOM]]") == "Hello World"

# Test chaining two functions
def test_chained_functions():
    def custom_func():
        return "Hello World"

    with context(CUSTOM=custom_func):
        assert splice("[[CUSTOM.UPPER]]") == "HELLO WORLD"


# Test a custom two parameter function
def test_two_parameter_function():
    def func_add_to_self(self, other):
        return self + other

    with context(APPEND=func_add_to_self, HELLO="Hello "):
        assert splice("[[HELLO.APPEND='World']]") == "Hello World"


# Test a custom filter that does nothing
def test_noop_filter():
    def custom_func(self):
        return self
    
    with context(NOOP=custom_func, HELLO="Hello World"):
        assert splice("[[HELLO.NOOP]]") == "Hello World"


# Test a custom function with a default parameter
def test_default_parameter_function():
    def concat_func(self, other="World"):
        return f"{self}{other}"

    with context(CONCAT=concat_func, HELLO="Hello "):
        assert splice("[[HELLO.CONCAT]]") == "Hello World"


# Test splitting a string
def test_split_string():
    # TODO: Fix this test
    with context(HELLO="Hello World"):
        assert splice("[[HELLO.SPLIT.ITEM=0]]") == "Hello"


# Test ADD function
def test_add_function():
    with context(HELLO="Hello ", WORLD="World"):
        assert splice("[[HELLO.ADD=WORLD]]") == "Hello WORLD"


# Test RSPLIT function
def test_rsplit_function():
    with context(HELLO="Hello World"):
        assert splice("[[HELLO.SPLIT.ITEM=1]]") == "World"


# Test extracting the second word from a string
def test_second_word():
    with context(HELLO="Hello Foo Bar"):
        assert splice("[[HELLO.WORD=1]]") == "Foo"


# Test EQ and NEQ
def test_eq():
    with context(A=1, B=2):
        assert splice("[[A.EQ=1]]") == "True"
        assert splice("[[A.EQ=2]]") == "False"
        assert splice("[[A.NE=1]]") == "False"
        assert splice("[[A.NE=2]]") == "True"


# Test LT and GT
def test_lt_gt():
    with context(A=1, B=2):
        assert splice("[[A.LT=2]]") == "True"
        assert splice("[[A.LT=1]]") == "False"
        assert splice("[[A.GT=2]]") == "False"
        assert splice("[[A.GT=0]]") == "True"


# Test LTE and GTE
def test_lte_gte():
    with context(A=1, B=2):
        assert splice("[[A.LE=2]]") == "True"
        assert splice("[[A.LE=1]]") == "True"
        assert splice("[[A.LE=0]]") == "False"
        assert splice("[[A.GE=2]]") == "False"
        assert splice("[[A.GE=1]]") == "True"
        assert splice("[[A.GE=0]]") == "True"


# Test arithmetic
def test_arithmetic():
    with context(A=1):
        assert splice("[[A.ADD=2]]") == "3"
        assert splice("[[A.SUB=2]]") == "-1"
        assert splice("[[A.MUL=2]]") == "2"
        assert splice("[[A.DIV=2]]") == "0.5"
        assert splice("[[A.MOD=2]]") == "1"


# Test executing a block of python code
def test_execute_python_code():
    code = """
def func():
    print("Hello [[USERPARAM]]")
func()
    """
    with context(USERPARAM="World"):
        assert execute(code) == "Hello World\n"


# Test executing a block of python code with a return value
# and using it as a condition
def test_execute_python_code_with_return():
    condition = "if '[[USERPARAM]]' == 'World': print('Yes')"
    with context(USERPARAM="World"):
        assert execute(condition) == "Yes\n"


# Test vanish and unvanish
def test_vanish_unvanish():
    text = "select * from [[TABLE]]"
    sql, sigils = vanish(text, pattern="?")
    assert sql == "select * from ?"
    assert sigils == ("[[TABLE]]",)
    assert unvanish(sql, sigils, pattern="?") == text



import os
import datetime
import pytest

from ..tools import *  # Module under test
from ..errors import SigilError, OnError
from ..sigils import Sigil
from ..contexts import context, global_context


def test_sigil_with_simple_context():
    with context(USER="arthexis"):
        assert splice("[[USER]]") == "arthexis"


def test_sigil_with_mapping_context():
    with context(ENV={"PROD": "localhost"}):
        assert splice("[[ENV='PROD']]") == "localhost"


def test_callable_no_param():
    with context(FUNC=lambda: "Test"):
        assert splice("[[FUNC]]") == "Test"


def test_class_static_attribute():
    class Entity:
        code = "Hello"

    with context(ENT=Entity()):
        assert splice("[[ENT.CODE]]") == "Hello"


def test_sigil_with_natural_index():
    with context(ENV=["hello", "world"]):
        assert splice("[[ENV=1]]") == "world"


def test_replace_missing_sigils_with_default():
    with context(USER="arthexis"):
        assert splice("[[NOT_USER]]", default="ERROR") == "ERROR"


def test_remove_missing_sigils():
    with context(USER="arthexis"):
        assert not splice("[[NOT_USER]]", on_error=OnError.REMOVE)


def test_attributes_casefold():
    class Env:
        def __init__(self, host):
            self.ssh_hostname = host

    hostname = "localhost"
    with context(ENV=Env(hostname)):
        assert splice("[[ENV.SSH_HOSTNAME]]") == hostname


def test_no_sigils_in_text():
    assert splice("No sigils") == "No sigils"


def test_call_lambda_same():
    with context(SAME=lambda arg: arg):
        assert splice("[[SAME='Test']]") == "Test"


def test_call_lambda_same_alt_quotes():
    with context(SAME=lambda arg: arg):
        assert splice('[[SAME="Test"]]') == "Test"


def test_call_lambda_reverse():
    with context(REVERSE=lambda arg: arg[::-1]):
        assert splice("[[REVERSE='Test']]") == "tseT"


def test_call_lambda_error():
    with context(DIVIDE_BY_ZERO=lambda arg: arg / 0):
        with pytest.raises(SigilError):
            splice("[[DIVIDE_BY_ZERO=1]]", on_error=OnError.RAISE)


def test_subitem_subscript():
    with context(A={"B": "C"}):
        assert splice("[[A.B]]") == "C"


def test_item_subscript_key_not_found():
    with context(A={"B": "C"}):
        with pytest.raises(SigilError):
            splice("[[A.C]]", on_error=OnError.RAISE)


def test_required_key_not_in_context():
    with context(USER="arthexis"):
        with pytest.raises(SigilError):
            splice("[[ENVXXX]]", on_error=OnError.RAISE)


def test_replace_duplicated():
    # TODO: Fix this test
    text = "User: [[U]], Manager: [[U]], Company: [[ORG]]"
    text, sigils = vanish(text, "X")
    assert sigils == ("[[U]]", "[[U]]", "[[ORG]]")
    assert text == "User: X, Manager: X, Company: X"

def test_resolve_simple_whitespace():
    with context({"ENV": "local"}):
        assert splice("[[ ENV ]]") == "local"


def test_resolve_dotted_whitespace():
    with context({"ENV": {"USER": "admin"}}):
        assert splice("[[ ENV  .  USER ]]") == "admin"


def test_resolve_recursive_one_level():
    with context(Y="[[X]]", X=10):
        assert splice("[[Y]]", recursion=1) == "10"


def test_sigil_helper_class():
    sigil = Sigil("Hello [[PLACE]]")
    assert sigil(PLACE="World") == "Hello World"


# Test converstion to json works with the standard library
# using the Sigil class
def test_json_conversion():
    import json
    with context(USER="arthexis"):
        sigil = Sigil("Hello [[USER]]")
        assert json.dumps(sigil) == '"Hello [[USER]]"'


# RJGO New functionatlity: using lists in the context
def test_item_subscript():
    with context(A=[1,2,3]):
        assert splice("[[A.ITEM=2]]") == "3"




# Test global_context
def test_global_context():
    global_context()["USER"] = "arthex1s"
    assert splice("[[USER]]") == "arthex1s"


# Test global_context
def test_global_context_set_key():
    global_context("USERA", "arthexe4s")
    assert splice("[[USERA]]") == "arthexe4s"




import logging
import io
from typing import Iterator, Mapping, Union, TextIO

import lark

logger = logging.getLogger(__name__)

GRAMMAR = r"""
    start       : sigil
    sigil       : "[" node ("." node)* "]"
    node        : CNAME ["=" arg]
                | arg
    arg         : sigil
                | "'" /[^']+/ "'"
                | "\"" /[^"]+/ "\""
                | integer
                | CNAME
                | "True"
                | "False"
                | "None"
                | null
    null        : "NULL"
    integer     : INT

    %import common.CNAME
    %import common.INT
    %import common.WS
    %ignore WS
"""

_parser = lark.Lark(GRAMMAR)
parse = _parser.parse


def spool(
        text: Union[str, TextIO],
        left: str = "[[",
        right: str = "]]"
) -> Iterator[str]:
    """Generator that extracts top-level sigils from text.

    :param text: The text to extract the sigils from.
    :param left: Left delimiter, defaults to "[[".
    :param right: Right delimiter, defaults to "]]".
    :return: An iterable of sigil strings.

    >>> # Extract simple sigils from text
    >>> list(spool("Connect to [[ENV.HOST]] as [[USER]]"))
    ['[[ENV.HOST]]', '[[USER]]']

    >>> # Nested sigils are allowed, only top level is extracted
    >>> list(spool("Environ [[SYS.ENV=[SRC]]"))
    ['[[SYS.ENV=[SRC]]']

    >>> # Malformed sigils are not extracted (no errors raised).
    >>> list(spool("Connect to [[ENV.HOST as USER"))
    []
    """
    assert len(left) == 2 and len(right) == 2, "Delimiters must be 2 characters long."
    buffer = []
    depth = 0
    _left, _right = left[0], right[0]  
    if isinstance(text, str): 
        text = io.StringIO(text) 
    while char := text.read(1):
        if char == _left:
            if text.read(1) == left[1]:
                depth += 1
        if depth > 0:
            buffer.append(char)
            if char == _right:
                if text.read(1) == right[1]:
                    depth -= 1
                if depth == 0:
                    token = ''.join(buffer)
                    if token.count("'") % 2 == 0 and token.count('"') % 2 == 0:
                        yield _left + token + _right
                    buffer.clear()


def _try_get_item(obj, key, index=None):
    try:
        found = obj[key]
        if index is not None:
            if isinstance(found, str) and ';' in found:
                found = found.split(';')
            return found[index]
        return found
    except (KeyError, TypeError):
        return None


def _try_call(func, *args) -> Union[None, str]:
    if callable(func):
        # logger.debug(f"Callable, try with {args=}.")
        args = tuple(arg for arg in args if arg is not None)
        if args: return func(*args)
        else: return func()
    else:
        # logger.debug(f"Non-callable {func}.")
        return None


def _try_manager(model):
    return getattr(model, "objects", None)


class SigilContextTransformer(lark.Transformer):

    def __init__(self, context: Mapping):
        super().__init__()
        self.ctx = context

    def _ctx_lookup(self, key):
        # logger.debug(f"Context lookup: {key}")
        try:
            value = self.ctx[key]
            # logger.debug(f"Found {type(value)} {value}")
            return value
        except KeyError as ex:
            # logger.debug(f"{key} not found.")
            raise ex

    @lark.v_args(inline=True)
    def sigil(self, *nodes):
        stack = list(nodes[::-1])
        # logger.debug(f"Resolving node stack {stack}.")
        # Process the first node (root)
        name, param = stack.pop()
        if isinstance(param, lark.Token): param = param.value
        if not (target := self._ctx_lookup(name)):
            # logger.debug(f"Root '{name}' not found in context.")
            return
        if manager := _try_manager(target):
            # We don't want to find managers after the first node
            # logger.debug("Found 'objects' attribute, treat as Django Model.")
            if param:
                # logger.debug(f"Search by pk={param}")
                if instance := manager.get(pk=param):
                    # logger.debug(f"Found instance with pk={param}.")
                    target = instance
                elif instance := manager.get_by_natural_key(param):
                    # logger.debug(f"Found instance with natural key: {param}.")
                    target = instance
            else:
                try:
                    # logger.debug("Lookahead into the next node.")
                    name, param = stack.pop()
                    if param: target = manager.get(**{name.casefold(): param})
                    else: stack.append((name, param))
                except IndexError:
                    pass  # Nothing to append back, pop failed
                except RuntimeError as ex:
                    stack.append((name, param))
        elif callable(target):
            # logger.debug("Root is callable, try to call.")
            target = _try_call(target, param)
        elif param and( value := _try_get_item(target, param)):
            # logger.debug(f"Lookup param '{param}' found.")
            target = value
        # Process additional nodes after the first (root)
        while stack:
            name, param = stack.pop()
            if isinstance(param, lark.Token): param = param.value
            # logger.debug(f"Consume {name=} {param=}.")
            if field := _try_get_item(target, name, param):
                # This finds items only (not attributes)
                # logger.debug(f"Field (exact) {name} found in {target}.")
                target = _try_call(field, param) or field
            elif field := getattr(target, name.casefold(), None):
                # Casefold is used to find Model fields, properties and methods
                # logger.debug(f"Field (casefold) {name} found in {target}.")
                target = _try_call(field, param) or field
            elif _filter := self._ctx_lookup(name):
                # We don't casefold filters (they are not Model fields)
                # logger.debug(f"Filter {name} found in context.")
                target = _try_call(_filter, target, param)
            # logger.debug("Unable to consume full sigil {name} for {target}.") 
        if isinstance(target, bytes):
            return target.decode()
        elif isinstance(target, (list, tuple)):
            target = ";".join(target)
        return target

    # Flatten nodes (otherwise a Tree is returned)
    @lark.v_args(inline=True)
    def node(self, key, arg=None):
        return [str(key), arg]

    # Flatten node args (otherwise a Tree is returned)
    @lark.v_args(inline=True)
    def arg(self, value=None):
        return value

    @lark.v_args(inline=True)
    def integer(self, value=None):
        return int(str(value))

    null = lambda self, _: ""


__all__ = ["spool", "SigilContextTransformer"]


from .tools import splice, spool, vanish, unvanish
from .contexts import context


class Sigil(str):
    """Encapsulates a string or stream that may contain sigils.
    """

    def __init__(self, original: str, **context):
        self.original = original
        self.context = context

    def __str__(self):
        """Resolve the sigil with splice and return the resulting text."""
        return splice(self.original, **self.context)
    
    def __repr__(self):
        """Return a original representation of the text with sigils."""
        return self.original

    def __call__(self, *args, **kwargs):
        """Send all args and kwargs to context, then resolve."""
        with context(*args, **kwargs):
            return splice(self.original, **self.context)
        
    def __iter__(self):
        """Iterate over the sigils in the text."""
        return iter(spool(self.original))
    
    @property
    def sigils(self):
        """Return a list of sigils in the text."""
        return list(self)
    
    def clean(self, pattern: str):
        """Replace all sigils in the text with another pattern."""
        return vanish(self.original, pattern)
    
    def dirty(self, sigils: list, pattern: str):
        """De-replace all patterns in the text with sigils."""
        return unvanish(self.original, sigils, pattern)  # type: ignore



__all__ = ["Sigil"]



import pytest
from dataclasses import dataclass

from ..tools import *  # Module under test
from ..errors import SigilError, OnError
from ..contexts import context

DATABASE = []


class Manager:

    @staticmethod
    def get(**kwargs):
        for obj in DATABASE:
            current = None
            for key, val in kwargs.items():
                if getattr(obj, key) == val:
                    current = obj
                else:
                    current = None
            if current:
                return current

    def get_by_natural_key(self, name):
        return self.get(name=name)


class Model:
    """Simulate a Django ORM model."""

    def save(self):
        DATABASE.append(self)

    objects = Manager()


@dataclass
class UserModel(Model):
    pk: int
    name: str
    alias: str


def test_model_has_objects_attribute():
    assert hasattr(UserModel, "objects")


def test_model_context_class():
    UserModel(pk=1, name="arthexis", alias="admin").save()
    with context(USR=UserModel):
        assert splice("[[USR]]", serializer=lambda x: x.__name__) == "UserModel"


def test_model_context_pk_attribute():
    UserModel(pk=1, name="arthexis", alias="admin").save()
    with context(USR=UserModel):
        assert splice("[[USR=1.NAME]]") == "arthexis"


def test_model_context_pk():
    UserModel(pk=2, name="arthexis", alias="admin").save()
    with context(USR=UserModel):
        assert splice("[[USR.NAME='arthexis'.PK]]") == "1"


def test_model_context_wrong_attribute():
    UserModel(pk=3, name="arthexis", alias="admin").save()
    with context(USR=UserModel):
        with pytest.raises(SigilError):
            assert splice("[[USR='admin'.NAME]]", on_error=OnError.RAISE)


def test_model_context_get_by_natural_key():
    UserModel(pk=3, name="arthexis", alias="admin").save()
    with context(USR=UserModel):
        assert splice("[[USR='arthexis'.ALIAS]]") == 'admin'


# Test serializing a model using json from the standard library
def test_model_json_serialization():
    import json
    UserModel(pk=3, name="arthexis", alias="admin").save()
    with context(USR=UserModel):
        assert json.loads(splice("[[USR='arthexis'.ALIAS]]", serializer=json.dumps)) == 'admin'




import os
import pytest

from ..parser import *  # Module under test
from ..tools import *  # Module under test

@pytest.mark.skip
def test_extract_slightly_nested():
    # Simple example
    sigil = "[[APP=[SYS.ENV.APP_NAME].MODULE.NAME]]"
    text = f"-- {sigil} --"
    assert set(spool(text)) == {sigil}

# TODO: Fix nested sigils
@pytest.mark.skip
def test_extract_single_deep_nested():
    # Very exagerated example
    sigil = "[[APP=[ENV=[REQUEST].USER]].OR=[ENV=[DEFAULT].USER].MODULE.NAME]]"
    text = f"-- {sigil} --"
    assert set(spool(text)) == {sigil}


def test_ignore_whitespace():
    sigil = "[[ ENV . HELP ]]"
    text = f"-- {sigil} --"
    assert set(spool(text)) == {sigil}


def test_extract_file_stream():
    # Chdir to test data directory
    os.chdir(os.path.dirname(__file__))
    with open("data/sample3.txt", "r") as fp:
        assert len(set(spool(fp))) == 3


# Test both kinds of quotes used in sub sigils
def test_extract_sub_sigils():
    sigils = ["[[ENV.HOST]]", "[[ENV.HELLO='World']]", "[[ENV.HELLO=\"World\"]]"]
    text = f"-- {sigils[0]} -- {sigils[1]} -- {sigils[2]} --"
    assert set(spool(text)) == set(sigils)
    

# Test that parsing fails when a sub sigil is not closed
def test_parse_sub_sigil_not_closed():
    # TODO: This should not be extracted
    sigil = "[[ENV.HELLO='World]]"
    text = f"-- {sigil} --"
    assert set(spool(text)) == set()



from enum import Enum


class SigilError(Exception):
    """Raised when a sigil cannot be resolved."""


# Rewrite the constants as an enum
class OnError(str, Enum):
    IGNORE = "ignore"
    RAISE = "raise"
    REMOVE = "remove"
    DEFAULT = "default"


__all__ = ["SigilError"]



import logging
import pathlib
import click    

from .tools import splice, execute
from .contexts import context as local_context

logging.basicConfig(level=logging.WARNING)


@click.command()
@click.argument("text", nargs=-1)
@click.option("--on-error", "-e", 
        default="ignore", type=click.Choice(["raise", "remove", "default", "ignore"]), 
        help="What to do when a sigil cannot be resolved. Default: ignore")
@click.option("--default", "-d", default="", 
        help="Default value for ignored sigils.")
@click.option("--verbose", "-v", count=True, help="Increase verbosity.")
@click.option("--interactive", "-i", is_flag=True, help="Enter interactive mode.")
@click.option("--file", "-f", help="Read text from given file.")
@click.option("--exec", "-x", is_flag=True, help="Execute the text after resolving sigils.")
@click.option("--context", "-c", multiple=True,
        help="Context to use for resolving sigils."
            "Can be a file or a key=value pair. Multiple values are allowed.")
def main(text, on_error, verbose, default, interactive, file, exec, context):
    """Resolve sigils found in the given text."""
    if verbose == 1:
        logging.basicConfig(level=logging.INFO)
    elif verbose > 1:
        logging.basicConfig(level=logging.DEBUG)

    if interactive:
        import code
        def readfunc(prompt):
            line = input(prompt)
            line = splice(line, on_error=on_error, default=default)
            return line
        code.interact(local=locals(), readfunc=readfunc)
        return
    
    if file:
        # Convert path to absolute path
        file = pathlib.Path(file).resolve()
        with open(file, 'r') as f:
            text = f.read()
    else:
        text = " ".join(text)
    with local_context(*context):
        if not exec:
            result = splice(text, on_error=on_error, default=default)
        else:
            result = execute(text, on_error=on_error, default=default)
        print(result)

main()



import io
import functools
import contextlib
import multiprocessing as mp
from typing import Union, Text, Iterator, Callable, Any, Optional, TextIO, Tuple, List

from . import contexts
from .parser import spool, parse, SigilContextTransformer
from .errors import SigilError, OnError


import logging
logger = logging.getLogger(__name__)


def splice(
        text: Union[str, TextIO],
        on_error: str = OnError.DEFAULT,
        default: Optional[str] = "",
        recursion: int = 6,
        serializer: Callable[[Any], str] = str,
) -> str:
    """
    Resolve all sigils found in text, using the local context, and return the
    resulting text. If the text contains no sigils, it's returned unchanged.

    :param text: The text containing sigils.
    :param on_error: What to do if a sigil cannot be resolved:
        OnError.DEFAULT: Replace the sigil with the default value (the default).
        OnError.IGNORE: Ignore the error and leave the sigil in the text output.
        OnError.RAISE: Raise a SigilError detailing the problem.
        OnError.REMOVE: Remove the sigil from the text output.
    :param serializer: Function used to serialize the sigil value, defaults to str.
    :param default: Value to use when the sigils resolves to None, defaults to "".
    :param max_recursion: If greater than zero, and the output of a resolved sigil
        contains other sigils, resolve them as well until no sigils remain or
        until max_recursion is reached (default is 6).

    >>> # Resolving sigils using context:
    >>> with context(ENV={"HOST": "localhost"}, USER="arthexis"}):
    >>>     resolve("Connect to [ENV.HOST] as [USER]")
    'Connect to localhost as arthexis'
    """

    if not isinstance(text, str):
        text = text.read()
    if not (sigils := set(spool(text))):
        return text  # Not an error, just do nothing
    results = []

    if len(sigils) > 1:
        with mp.Pool() as pool:
            results = pool.starmap(splice, [
                (sigil, on_error, default, recursion)
                for sigil in sigils
            ])
    else:
        sigil = sigils.pop()
        try:
            # By using a lark transformer, we parse and resolve
            # each sigil in isolation and in a single pass
            tree = parse(sigil[1:-1])
            transformer = SigilContextTransformer(contexts._threadlocal.context)
            value = transformer.transform(tree).children[0]
            # logger.debug("Sigil '%s' resolved to '%s'.", sigil, value)
            if value is not None:
                fragment = serializer(value)
                if recursion > 0:
                    fragment = splice(
                        fragment,
                        on_error,
                        default,
                        recursion=(recursion - 1)
                    )
                text = text.replace(sigil, fragment)
        except Exception as ex:
            if on_error == OnError.RAISE:
                raise SigilError(sigil) from ex
            elif on_error == OnError.REMOVE:
                text = text.replace(sigil, "")
            elif on_error == OnError.DEFAULT:
                text = text.replace(sigil, str(default))

    if results:
        for sigil, result in zip(sigils, results):
            text = text.replace(sigil, result)
    return text


# Resolve is a version of splice that defaults to recursion = 0
resolve = functools.partial(splice, recursion=0)


def vanish(
        text: str,
        pattern: Union[Text, Iterator],
) -> tuple[str, tuple[str]]:
    """
    Replace all sigils in the text with another pattern.
    Returns the replaced text and a list of sigils in found order.
    This will not resolve the sigils by default.

    :param text: The text with the sigils to be replaced.
    :param pattern: A text or iterator used to replace the sigils with.
    :return: A tuple of: replaced text, list of sigil strings.

    >>> # Protect against SQL injection.
    >>> replace("select * from users where username = [USER]", "?")
    ('select * from users where username = ?', ['[USER]'])
    """

    sigils = list(spool(text))
    _iter = (iter(pattern) if isinstance(pattern, Iterator) 
             else iter(pattern * len(sigils)))
    for sigil in set(sigils):
        text = (text.replace(sigil, str(next(_iter)) or sigil))
    return text, tuple(sigils)


def unvanish(
        text: str,
        sigils: Tuple[str] | List[str],
        pattern: Union[Text, Iterator],
) -> str:
    """
    De-replace all patterns in the text with sigils.

    :param text: The text with the pattern to be replaced.
    :param sigils: A list of sigils to be replaced in.
    :param pattern: A text or iterator used to replace the patterns with.
    :return: A tuple of: de-replaced text.
    """

    _iter = (iter(pattern) if isinstance(pattern, Iterator) 
             else iter(pattern * len(sigils)))
    for sigil in set(sigils):
        text = (text.replace(str(next(_iter)) or sigil, sigil))
    return text


def execute(
        code: str,
        on_error: str = OnError.DEFAULT,
        default: Optional[str] = "",
        recursion_limit: int = 6,
        _locals: Optional[dict[str, Any]] = None,
        _globals: Optional[dict[str, Any]] = None,
) -> Union[str, None]:
    """Execute a Python code block after resolving any sigils appearing in it.
    If the executed code block prints to stdout, it's output is returned. 
    If the code prints to stderr, a SigilError is raised.
    If the code block looks like an expression, it's value is returned.
    If the code block looks like a statement, None is returned.
    If the code block contains no sigils, it's not executed and None is returned.
    """
    if not set(spool(code)):
        return None
    _code = splice(code, on_error=on_error, default=default, 
                    recursion=recursion_limit)
    with contextlib.redirect_stdout(io.StringIO()) as f:
        with contextlib.redirect_stderr(io.StringIO()) as err:
            exec(_code, _globals, _locals)
            if err.getvalue():
                raise SigilError(err.getvalue())
    return f.getvalue()


__all__ = ["spool", "splice", "execute", "vanish", "unvanish", "resolve"]



import os
import sys
import uuid
import math
import json
import random
import base64
import socket
import urllib.parse
import collections
import contextlib
import threading
from typing import Generator, Optional, Any

import logging
logger = logging.getLogger(__name__)

try:
    if not os.environ.get("DJANGO_SETTINGS_MODULE"):
        logger.debug("No DJANGO_SETTINGS_MODULE set, using datetime.datetime")
        raise ImportError
    from django.utils import timezone as datetime
except ImportError:
    from datetime import datetime


class System:
    """Used for the SYS default context."""

    # TODO: Check user privileges before accessing SYS 
    # TODO: Missing tests

    class _Env:
        def __getitem__(self, item):
            return os.getenv(item.upper())
        
        def __str__(self):
            return str(','.join(os.environ.keys()))

    _env = _Env()

    @property
    def env(self): return System._env
            
    @property
    def args(self): return sys.argv[1:]

    @property
    def now(self): return datetime.now()

    @property
    def today(self): return datetime.now().date()

    @property
    def time(self): return datetime.now().time()

    @property
    def uuid(self): return str(uuid.uuid4()).replace('-', '')

    @property
    def rng(self): return random.random()

    @property
    def pid(self): return os.getpid()

    @property   
    def pi(self): return math.pi

    @property
    def os(self): return os.name

    @property
    def arch(self): return sys.platform

    @property
    def host(self): return socket.gethostname()

    @property
    def ip(self): return socket.gethostbyname(socket.gethostname())

    @property
    def user(self): return os.getlogin()

    @property
    def home(self): return os.path.expanduser("~")

    @property
    def python(self): return sys.executable

    @property
    def py_ver(self): return sys.version

    @property
    def sig_ver(self): 
        # TODO: Missing tests
        with open(os.path.join(os.path.dirname(__file__), "..", "pyproject.toml")) as f:
            return json.loads(f.read())["project"]["version"]
        
    @property
    def pwd(self): return os.getcwd()

    @property
    def cwd(self): return os.getcwd()
        
    @property
    def tmp(self): return os.path.join(os.getcwd(), "tmp")

class ThreadLocal(threading.local):
    def __init__(self):
        # These will be the default filters and objects
        self.context = collections.ChainMap({
            "SYS": System(),
            "NUM": lambda x: float(x) if "." in x else int(x),
            "UPPER": lambda x: str(x).upper(),
            "LOWER": lambda x: str(x).lower(),
            "TRIM": lambda x: str(x).strip(),
            "STRIP": lambda x: str(x).strip(),
            "SLUG": lambda x: str(x).replace(" ", "-"),
            "ZFILL": lambda x, y: str(x).zfill(y),
            "F": lambda x, *args, **kwargs: x.format(*args, **kwargs),
            "STYLE": lambda x: dict(i.split(":") for i in x.split(";") if i),
            "JOIN": lambda o, s: (s or "").join(str(i) for i in o),
            "SPLIT": lambda x, y: str(x).split(y),
            "OR": lambda x, y: x or y,
            "AND": lambda x, y: x and y,
            "NOT": lambda x: not x,
            "BOOL": lambda x: bool(x),
            "INT": lambda x: int(x),
            "FLOAT": lambda x: float(x),
            "TUPLE": lambda x: tuple(x),
            "LIST": lambda x: list(x),
            "SET": lambda x: set(x),
            "JSON": lambda x: json.dumps(x),
            "B64": lambda x: base64.b64encode(x),
            "B64D": lambda x: base64.b64decode(x),
            "URL": lambda x: urllib.parse.quote(x),
            "URLD": lambda x: urllib.parse.unquote(x),
            "LEN": lambda x: len(x),
            "REV": lambda x: x[::-1],
            "SORT": lambda x: sorted(x),
            "ITEM": lambda x, y: x[y],
            "KEY": lambda x, y: x[y],
            "ATTR": lambda x, y: getattr(x, y),
            "ANY": lambda x: any(x),
            "ALL": lambda x: all(x),
            "NONE": lambda x: not any(x),
            "SUM": lambda x: sum(x),
            "MIN": lambda x: min(x),
            "MAX": lambda x: max(x),
            "AVG": lambda x: sum(x) / len(x),
            "ABS": lambda x: abs(x),
            "ROUND": lambda x, y: round(x, y or 0),
            "CEIL": lambda x: int(x) + 1,
            "FLOOR": lambda x: int(x),
            "TRUNC": lambda x: int(x),
            "MOD": lambda x, y: x % y,
            "ADD": lambda x, y: x + y,
            "SUB": lambda x, y: x - y,
            "MUL": lambda x, y: x * y,
            "DIV": lambda x, y: x / y,
            "FDIV": lambda x, y: x // y,
            "EQ": lambda x, y: str(x) == str(y),
            "NE": lambda x, y: str(x) != str(y),
            "LT": lambda x, y: x < y,
            "LE": lambda x, y: x <= y,
            "GT": lambda x, y: x > y,
            "GE": lambda x, y: x >= y,
            "IN": lambda x, y: x in y,
            "CONTAINS": lambda x, y: y in x,
            "FIRST": lambda x: x[0],
            "LAST": lambda x: x[-1],
            "HEAD": lambda x, y: x[:y],
            "TAIL": lambda x, y: x[-y:],
            "TYPE": lambda x: type(x).__name__,
            "FLAT": lambda x: [i for j in x for i in j],
            "ESC": lambda x: x.replace("\\", "\\\\").replace('"', '\\"'),
            "UNIQ": lambda x: list(set(x)),
            "ZIP": lambda x, y: list(zip(x, y)),
            "SIG": lambda x: f"[[{x}]]",
            "WORD": lambda x, y: x.split()[y],
        })
        # Add default context sources and thread local variables here


_threadlocal = ThreadLocal()


# Function that recursively changes the keys of a dictionary
# to uppercase. This is used to make the context dictionary uniform.
def _upper_keys(d):
    if isinstance(d, dict):
        return {k.upper(): _upper_keys(v) for k, v in d.items()}
    else:
        return d


def _load_context_file(filename):
    filename = os.path.abspath(filename)
    if filename.endswith(".toml"):
        import tomllib
        with open(filename, 'r') as f:
            return _upper_keys(tomllib.loads(f.read()))
    elif filename.endswith(".json"):
        import json
        with open(filename, 'r') as f:
            return _upper_keys(json.loads(f.read()))
    elif filename.endswith(".yaml") or filename.endswith(".yml"):
        import yaml
        with open(filename, 'r') as f:
            return _upper_keys(yaml.safe_load(f.read()))
    elif filename.endswith(".ini"):
        import configparser
        config = configparser.ConfigParser()
        config.read(filename)
        file_data = {}
        for section in config.sections():
            file_data[section.upper()] = _upper_keys(dict(config.items(section)))
        return file_data
    elif filename.endswith(".env"):
        import dotenv
        return _upper_keys(dotenv.dotenv_values(filename))
    else:
        raise ValueError(f"Unknown file type: {filename}")


@contextlib.contextmanager
def context(*args, **kwargs) -> Generator[collections.ChainMap, None, None]:
    """Context manager that updates the local context used by sigils.

    :param args: A tuple of context sources.
    :param kwargs: A mapping of context keys to values or functions.

    >>> # Add to context using kwargs
    >>> with context(TEXT="hello world") as ctx:
    >>>     assert ctx["TEXT"] == "hello world"
    """
    global _threadlocal
    _threadlocal.context = _threadlocal.context.new_child(kwargs)

    for arg in args:
        if isinstance(arg, dict):
            _threadlocal.context = _threadlocal.context.new_child(arg)
        elif callable(arg):
            _threadlocal.context = _threadlocal.context.new_child(arg())
        elif isinstance(arg, str):
            if "=" in arg:
                key, value = arg.split("=", 1)
                _threadlocal.context[str(key).upper()] = value  # type: ignore
            else:
                file_data = _load_context_file(arg)
                _threadlocal.context = _threadlocal.context.new_child(file_data)
            
    yield _threadlocal.context
    _threadlocal.context = _threadlocal.context.parents


def global_context(key: Optional[str] = None, value: Any = None) -> Any:
    """Get or set a global context value.
    You should normally use the context() function instead.
    
    :param key: The key to get or set.
    :param value: The value to set.
    :return: The value of the key or the entire context.

    >>> # Get the entire context
    >>> global_context()
    >>> # Get a value from the context
    >>> global_context("TEXT")
    >>> # Set a value in the context
    >>> global_context("TEXT", "hello world")
    """
    global _threadlocal
    if key and value is None:
        return _threadlocal.context[key]
    elif key and value is not None:
        _threadlocal.context[key] = value
    return _threadlocal.context


__all__ = ["context", "global_context"]	



import subprocess as sp


# Test calling from the command line
def test_cli_error_missing_context():
    result = sp.run(["sigils", "Hello [[WORLD]]", "-e", ""], capture_output=True)
    assert result.stdout == b""
    assert result.stderr != b""


# Test calling from the command line with a custom function
def test_cli_custom_function():
    result = sp.run(
        ["sigils", "[[SYS.ENV.PATH]]", "--on-error", "ignore"], 
        capture_output=True)
    assert result.stdout not in (b"", b"None", b"[[SYS.ENV.PATH]]")
    assert result.stderr == b""
